% A set of x-y data are fit (single independent variable). x extends to 60 but the spline is
% determined beyond the data.  The data are hard to fit with conventional polynomials
% since there are large and variable second derivatives for small x and
% smaller second derivatives for larger x.  A range of damping values
% is used - from overdamped (essentially a linear fit to data) to
% underdamped (over fitting data) - notice that damping=1 achieves a
% reduced chi square near 1.  Note also that the fit in the extrapolated
% region has zero second derivatives. Regularization in the absence of data
% controls the nature of the fit.

% the function spdft:  first input: independent variables (either cell or matrix)
%                      second input: dependent variable (column of "observations")
%                      third input: uncertainties (empty for equal weight)
%                      fourth input: cell containing control points for independent variables
%                      fifth input: damping (as a vector as long as number of independent variables)
%                      sixth input: number of regularization points per knot interval(as a vector as long as number of independent variables)
%                      7th input: order of splines
%                      8th input: derivative to be made small for regularization


% Here are the data to fit:
data2fit=XY2fit_data;
X1=data2fit(:,1);
Y1=data2fit(:,2);
uncert=.05;  % set constant data uncertainty - can be vector or empty

% Below here, the spline fitting parameters are described. One must:
%      1. define control points, 
%      2. determine how many regularization points per control point (1 is a typical value)
%      3. specify the order of the spline, 
%      4. specify what derivative to minimize,
%      5. specify what damping factor to use

% for small control point spacings the fit should be insensitive to spacing choice.  
% If the spacings are too wide, the spline will not have sufficient flexibility to fit the data
dx=.5;  % try .1 .5 1 2 5 10 
Xc=0:dx:100;  % the spline control points are set

RegFac=1;  % chose amount of over-sampling of regularization (empty for no regularization) 
%              1 for 1 regularization point per control point

Order=4;  % increasing the spline order can compensate for wider control point spacing 
%           (try orders between 4 (minimum for smoothing) and 10 with larger or smaller dx values)

% here a smooth fit requires minimizing the second derivative of the fit:
mdrv=2;

% try a range of damping factors - near 1 is always a place to start
lam=[ 500 50 5 1 .1 .001 ];


scrsz=get(0,'ScreenSize');
figsz=[1 .8*scrsz(4) 2*scrsz(3)/4 .8*scrsz(4)];
h=figure('Name','One Independent Variable','Position',figsz);

for i=1:length(lam)
    subplot(3,2,i)
% here is the call to the spline fitting function:
    sp1=spdft(X1,Y1,uncert,Xc,lam(i),RegFac,Order,mdrv,[]);
% spdft puts fit statistics into the returned spline structure - pull them
% out and create a text for the plots:
txt=sprintf('lambda= %6.3f \nrms misfit = %5.2f \nchisqr = %3.1f',lam(i),sp1.Data.rms,sp1.Data.chisqr);

plot(X1,Y1,'ko','MarkerFaceColor','w')
hold on
fnplt(sp1) % fnplt is a MATLAB utility for quickly plotting spline representations
text(40,.9,txt,'FontSize',14)
axis([0 100 -5 2 ])
hold off

end
