% This script is associated with the section of "Local Basis Function
% Representations of Thermodynamic Surfaces" by J.M. Brown labeled:
%
% "Representation and  Modification of IAPWS using Local Basis Functions "

% SECTION 1
% in this section a spline is constructed that represents IAPWS extending to 100 GPa and 20000 K
% the spline in fnIAPWSval had some adjustments for the edges - I used a
% different mask for its creation than for its evaluation - some
% edge effects are cut out by moving the mask to higher T at high P

% load a saved grid of IAPWS values (here to save time)
load IAPWS_grids
%rename variables to use:
P=IAPWS_TPa.P;
T=IAPWS_TPa.T;
rho=IAPWS_TPa.rho;
vel=IAPWS_TPa.vel;
mask=IAPWS_TPa.mask;
Cp=IAPWS_TPa.Cp;
G=IAPWS_TPa.G;

% The IAPWS formulation can be called to create the grid.  This
% takes considerably more time - do not uncomment unless you are willing to
% wait.
%         [rho,P,H,S,Cv,Cp,alpha,vel,E,G,gamma,Kt]=IAPWS95({P,T},'P'); 

% chose the region to include in the spline:
idp=find(P<=1e5);
idt=find(T<1e4);

% put P and T into a cell structure and set values to fit
PTf={P(idp),T(idt)}; 
Gf=G(idp,idt);
rhof=rho(idp,idt);
Cpf=Cp(idp,idt);
velf=vel(idp,idt);

% Things to note: IAPWS does not give accurate results in the "non-physical" region.  
% Thus, an inversion for the spline needs to include only grid points in the physical region.
% The saved variable "mask" is a grid of ones and NaNs to deliniate the  
% non-physical regions. The function mk_mask4Gspline clips out the "wrong areas" by making the variable "mask".
mask=mk_mask4Gspline(PTf); 

% things to adjust:
wt=1e0*[3e7 1e3]; % relative weight of densities and Cp's in fit
lam=1e0*[1 1]; % regularization damping
mdrv=[4 4];    % derivatives of G in P and T to be made small
ordr=[6 6];   % order of splines to use (at least 6 for G and larger to allow more flexibility
nreg=1;  % # of regularization points per control ponts
normflg=0; % rescale data if =1
% PTc=PTf;
% PTc{1}=PTc{1}-1;
% PTc{1}(end)=PTf{1}(end);
% here is the inversion for a G spline: need to fiddle to get the best results,
 tic
 sp_G=spgft(Gf(:,:),rhof,Cpf,wt,PTf,PTf,mask,lam,mdrv,ordr,nreg,normflg);  % use control points equal to gridded input
 toc
%
% Test the results 
[rhoc,velc,Gc,Cpc]=fnGval(sp_G,PTf);
% uncomment the following to see what the saved spline does
%[rhoc,velc,G,Cpc]=fnIAPWSval(PTf);
devsvel=real(1e6*(velc-velf)./velc);
devsrho=real(1e6*(rhoc-rhof)./rhoc);
devsCp=real(1e6*(Cpc-Cpf)./Cpc);
id=find(not(isnan(mask)) & isfinite(devsvel));
devs=sum(abs([devsrho(id);devsCp(id);devsvel(id)]));
[~,~,~,~,~,~,~,~,~,~,mask]=fnIAPWSval(PTf);
Pmax=PTf{1}(end);
Tmax=PTf{2}(end);
%
% here the spline for Gibbs energy is compared to the original IAPWS-95
% formulation.  Two figures with different PT ranges
%
figure(100)
subplot(321)
therm_surf(PTf,eps+devsrho,mask,'density deviations (ppm)',[0 1e3  200 1000 -1 1])
title('Deviations at Low Pressures and Temperatures')
subplot(323)
therm_surf(PTf,eps+devsvel,mask,'sound speed deviations (ppm)',[0 1e3  200 1000 -100 100])
subplot(325)
therm_surf(PTf,eps+devsCp,mask,'Cp deviations (ppm)',[0 1e3  200 1000 -1 1])

subplot(322)
therm_surf(PTf,eps+devsrho,mask,'density deviations (ppm)',[0 Pmax  200 Tmax -10 10])
title('Deviations over Entire Range of Pressures and Temperatures')
subplot(324)
therm_surf(PTf,eps+real(devsvel),mask,'sound speed deviations (ppm)',[0 Pmax 200 Tmax -100 100])
subplot(326)
therm_surf(PTf,eps+devsCp,mask,'Cp deviations (ppm)',[0 Pmax 200 Tmax -10 10])

%%
% SECTION TWO
% In this section differences are calculated between the spline representaton of IAPWS and
% the original formulation.  Values calculated at random locations
warning off  % warnings are issued by fnIAPWSval when points are out of the area
tic
% define the range of P and T to test
Trange=[250 1e4];
Prange=[0.1 1e5];
npt=2000; % number of points to try
% generate a random log distribution of pressures and temperatures:
T = 10.^(log10(Trange(1)) + (log10(Trange(2))-log10(Trange(1))).*rand(npt,1));
P = 10.^(log10(Prange(1)) + (log10(Prange(2))-log10(Prange(1))).*rand(npt,1));

% assemble as a matrix with P in first column and T in second column
PTs=[P(:) T(:)];

% get thermodynamic values using the spline representation saved with fnIAPWSval
[rho2,vel2,G2,Cp2,~,~,~,~,~,~,mask]=fnIAPWSval(PTs);

%some of the points will be in the excluded regions - remove them
id=find(not(isnan(mask)));
rho2=rho2(id);
vel2=vel2(id);
Cp2=Cp2(id);
%
% evaluate thermodynamic properties using the full IAPWS representation
[rho,vel,G,Cp]=IAPWS95(PTs(id,:),'P');

% calculate absolute fractional differences
devsrho=1e6*abs((rho-rho2)./rho2);
devsCp=1e6*abs((Cp-Cp2)./Cp);
devsvel=1e6*abs((vel-vel2)./vel);

disp(' Densities deviations (ppm): average    spread   maximum')
disp(sprintf('                              %6.3g   %6.3g    %6.3g',  mean(devsrho), std(devsrho) ,max(devsrho)))
disp(' Cp deviations (ppm): average    spread   maximum')
disp(sprintf('                       %6.3g   %6.3g   %6.3g', mean(devsCp), std(devsCp) ,max(devsCp)))
disp(' sound speed deviations (ppm): average    spread   maximum')
disp(sprintf('                               %6.3g    %6.3g    %6.3g', mean(devsvel), std(devsvel), max(devsvel)))


toc
% here is typical output from this sections
% If a point falls on the metastable boundary, values can show larger
% deviations - try running again.
%
%  Densities deviations (ppm): average    spread   maximum
%                               0.000685   0.00557     0.095
%  Cp deviations (ppm): average    spread   maximum
%                        0.0176    0.131     3.16
%  sound speed deviations (ppm): average    spread   maximum
%                                 0.107      1.27      37.6
% 
% Elapsed time is 6.055752 seconds.

%%
% SECTION THREE
% this section integrates IAPWS sound speeds
% to 100 GPa to confirm that the resulting EOS is essentially the same as
% the base EOS.  This is a test that the numerical framework is adequate.

% define P and T for "data"
Pd=[0.1:5:95.1 logspace(2,log10(1e5),300)]; % log grid of pressure points above 100 MPa
Pd(end)=1e5; % force last point to be exactly specified (the logspace command differs by machine precision)
Td=logspace(log10(250),log10(1.e4),200); % log grid of temperature points
Td(1)=250;Td(end)=1e4;
idhp=find(Pd>=1e3);  % later only the >1 GPa part will be modified so track the indeces of the two regions
idlp=find(Pd<1e3);

% define the Control Points for the splines: Fewer control points than data
% points and control points spanning the full range of data points
Pc=[0.:10:90. logspace(2,log10(1e5),200)];
Tc=logspace(log10(240),log10(1.e4),100);
% avoid round off issues with the logspace function - inforce the value at
% the beginning and end
Tc(1)=240;Tc(end)=1.e4;

PTd={Pd,Td};      % put Pd and Td into a cell structure
PTc={Pc,Tc};      % put Pc and Tc into a cell structure
PTdhp={Pd(idhp),Td};  % also assemble cell with onlyt the pressure points above 1000 MPa

% get "data" based on spline representation of IAPWS
[rho,vel,G,Cp,~,~,~,~,~,~,mask]=fnIAPWSval(PTd);

% The alternative is to use the original IAPWS formulation.  This is
% much slower to execute and is commented out here
%[rho,P,H,S,Cv,Cp,alpha,vel,E,G,gamma,Kt]=IAPWS95(PTd,'P');  % This does not
%return a mask for the areas not well represented.
%

lamv=16e-1; % degree of smoothing of the sound speed surface
% here is the spline fit for sound speeds
spv=spdft(PTd,vel,1e-8*vel,PTc,lamv*[1 1],[1 1],[4 4],[2 2],mask);
%
% evaluate sound speeds on the "data" grid - advantage of the spline is
% that it extrapolates reasonably into the metastable reagion where IAPWS
% does not work
vspline=fnval(spv,PTd);
% the following is a check that the sound speed surface just created is acceptable
devs=1e6*mask.*(vspline-vel)./vel;
%
figure(200)
subplot(212)
therm_surf(PTd,eps+devs,mask,'Vel Devs (ppm)',[0 1e5 250 1e4 -50 50])
title('This Figure provides Validation of the Local Basis Function Representation for Sound Speeds')
subplot(211)
therm_surf(PTd,vspline,mask,'Velocities (km/s)')
%
% here is the integrations of sound speeds to high pressure for the test
% equation of state
lam_eqst=.1e0;  % degree of smoothing of volumes vs temperature for eqst integration (arbitrary units and magnitude)
pflg=0; % set to 1 to see results at each pressure step
lam_fac=1.154; % factor to increase smoothing with increasing pressure - 1 for no change, 1.1 for 10% per step in P
lam_fac=1.2;

lam_eqst=1e0;  % degree of smoothing of volumes vs temperature for eqst integration (arbitrary units and magnitude)
pflg=0; % set to 1 to see results at each pressure step
lam_fac=1.15; % factor to increase smoothing with increasing pressure - 1 for no change, 1.1 for 10% per step in P
lam_fac=1.16;

% run integration
[rho2,Cp2,G2]=eqst(PTdhp,Tc,rho(idhp(1),:),Cp(idhp(1),:),vspline(idhp,:),G(idhp(1),:),mask(idhp,:),lam_eqst,lam_fac,pflg);
%[rho2,Cp2,G2]=eqst(PTd,Tc,rho(1,:),Cp(1,:),vspline(:,:),G(1,:),mask(:,:),lam_eqst,lam_fac,pflg);
%
% convert results to spline in Gibbs energy
% attach the EOS values below 1000 MPa
rho3=rho;
rho3(idhp,:)=rho2;
Cp3=Cp;
Cp3(idhp,:)=Cp2;
G3=G;
G3(idhp,:)=G2;
%
% some parameters needed for spgft:
wt=1e0*[1e3 1e0]; % relative fitting weight for rho and Cp
lamG=1e-1*[1 1]; % regularization damping
mdrv=[4 4];    % derivatives in P and T to be made small
ordr=[6 6];   % order of splines to use (at least 6 for G and larger to allow more flexibility
nreg=1;  % # of regularization points per control ponts
normflg=0; % rescale data if =1

% here is the inversion for a G spline: need to fiddle to get the best results,
sp_G_IAPWS=spgft(G3(1,:),rho3,Cp3,wt,PTd,PTc,mask,lamG,mdrv,ordr,nreg,normflg); 

%
%How well does the G spline work? - compare to the original IAPWS output
[rhoc,velc,Gc,Cpc]=fnGval(sp_G_IAPWS,PTd);
devsrho=1e2*(rho-rhoc)./rho;
devsCp=1e2*(Cp-Cpc)./Cp;
devsvel=1e2*(vel-velc)./vel;
devsG=eps+(G-Gc);
%
figure(600)
subplot(221)
therm_surf(PTd,eps+devsrho,mask,'Density Deviations (%)',[0 1e5 240 1e4 -1 1])
subplot(222)
therm_surf(PTd,eps+devsCp,mask,'Cp Deviations (%)',[0 1e5 240 1e4 -1 1])
subplot(223)
therm_surf(PTd,devsvel,mask,'Velocity Deviations (%)',[0 1e5 240 1e4 -1 1])
%therm_surf(PTd,devsvel,mask,'Velocity Deviations (%)',[0 2e3 240 1e3 -.5 .5])
subplot(224)
therm_surf(PTd,devsG,mask,'G Deviations (J/kg)')


figure(500)
subplot(221)
therm_surf(PTd,rho3,mask,'Density (kg/m^3)')
subplot(222)
therm_surf(PTd,Cp3,mask,'Cp (J kg^-^1 K^-^1)',[0 1e5 240 1e4 2000 5000])
subplot(223)
therm_surf(PTd,velc,mask,'Velocity  (m/s)',[0 1e5 240 1e4 1000 12000])
subplot(224)
therm_surf(PTd,Gc,mask,'G  (J/kg)')


%%
% SECTION FOUR
% this section integrates sound speeds modified to better match experiments
% at high pressure

% set the base IAPWS values
[rho,vel,G,Cp,~,~,~,~,~,~,mask]=fnIAPWSval(PTd);

% Create a surface for sound speed deviations from IAPWS (cor in percent)
p=[1e3 5e3 1e4   2e4 4e4 5e4 7.e4 8.5e4 1e5];
t=[240 5e3 1e4];
%     1e3    5e3           1e4         2e4           4e4     5e4      7e4         8.5e4     1e5
cor=[0 0 0; 3.3 3.3 3.3;  4. 4. 4.;    4.2 4.2 4.2; 4.5 4.5 4.5;  2 2 2;   -1 -1 -1;  -4 -4 -4;   -7 -7 -7 ]; %selection 4
 [pm,tm]=meshgrid(p,t);
 [pc,tc]=meshgrid(Pd(idhp),Td);
 corP=interp2(pm,tm,cor',pc,tc,'spline');
 
% create a surface of perturbed sound speeds and fit a spline to that
velp=vel;
velp(idhp,:)=(1-1e-2*corP'.^1).*vel(idhp,:);

spvm=spdft(PTd,velp,1e-3*velp,PTc,lamv*[10 1],[1 1],[4 4],[2 2],mask);

vsplinem=fnval(spvm,{Pd,Td});
devs=1e2*mask.*(vsplinem-vel)./vel;

figure(800)
subplot(212)
therm_surf(PTd,devs,mask,'Vel Devs (p%)')
subplot(211)
therm_surf(PTd,vsplinem,mask,'Velocities (km/s)')

% integrate the perturbed sound speeds

% run integration
[rho2,Cp2,G2]=eqst(PTdhp,Tc,rho(idhp(1),:),Cp(idhp(1),:),vsplinem(idhp,:),G(idhp(1),:),mask(idhp,:),lam_eqst,lam_fac,pflg);
% convert results to spline in Gibbs energy
% add back the EOS below 100 MPa
rho3=rho;
rho3(idhp,:)=rho2;
Cp3=Cp;
Cp3(idhp,:)=Cp2;
G3=G;
G3(idhp,:)=G2;
%
% use set up from control integration - see above

% here is the inversion for a G spline: need to fiddle to get the best results,
sp_G_IAPWS_modified=spgft(G3(1,:),rho3,Cp3,wt,PTd,PTc,mask,lamG,mdrv,ordr,nreg,normflg);

%How well does the G spline work? - compare to the original IAPWS output
[rhoc,velc,Gc,Cpc]=fnGval(sp_G_IAPWS_modified,PTd);
devsrho=1e2*(rho-rhoc)./rho;
devsCp=1e2*(Cp-Cpc)./Cp;
devsvel=1e2*(vel-velc)./vel;
devsG=(G-Gc);
%
figure(840)
subplot(221)
therm_surf(PTd,devsrho,mask,'Density Deviations (%)') %,[0 1e5 240 1e4 -10 10])
subplot(222)
therm_surf(PTd,devsCp,mask,'Cp Deviations (%)',[0 1e5 3e2 1e4 -2 5])
subplot(223)
therm_surf(PTd,devsvel,mask,'Velocity Deviations (%)',[0 1e5 240 1e4 -10 10])
subplot(224)
therm_surf(PTd,devsG,mask,'G Deviations (J/kg)')

figure(850)
subplot(221)
therm_surf(PTd,rhoc,mask,'Density (kg/m^3)')
subplot(222)
therm_surf(PTd,Cpc,mask,'Cp (J kg^-^1 K^-^1)',[0 1e5 240 1e4 2000 5000])
subplot(223)
%therm_surf(PTd,velc,mask,'Velocity  (m/s)',[0 1e5 240 1e4 500 13000])
therm_surf(PTd,velc,mask,'Velocity) (m/s)',[0 2e3 240 1e3 500 3000])
subplot(224)
therm_surf(PTd,Gc,mask,'G  (J/kg)',[0 1e5 240 1e4 -1e8 4e7])
%
% plot model and deviations in low pressure regime:
figure(330)
subplot(221)
therm_surf(PTd,rhoc,mask,'Density (kg/m^3)',[0 2e3 240 1200 00 1400])
subplot(222)
therm_surf(PTd,Cpc,mask,'Cp (J kg^-^1 K^-^1)',[0 2e3 240 1200 2000 6000])
subplot(223)
therm_surf(PTd,velc,mask,'Velocity  (m/s)',[0 2e3 240 1200 500 4000])
subplot(224)
therm_surf(PTd,Gc,mask,'G  (J/kg)',[0 2e3 240 1200 -4e6 2e6])

%
figure(340)
subplot(221)
therm_surf(PTd,devsrho,mask,'Density Deviations (%)',[0 2e3 240 1200 -5 5])
subplot(222)
therm_surf(PTd,devsCp,mask,'Cp Deviations (%)',[0 2e3 240 1200 -2 4])
subplot(223)
therm_surf(PTd,devsvel,mask,'Velocity Deviations (%)',[0 2e3 240 1200 -5 5])
subplot(224)
therm_surf(PTd,devsG,mask,'G Deviations (J/kg)',[0 2e3 240 1200 -1e3 2e3])


%
% Compare high pressure (shock wave and sound speeds) data with model
load WaterDataSets
load IAPWS_sp_strct
% Lyzenga data
%      P    rho  T (K)
dataL=[48.9 2.26 3280
       58.5  2.34 3830
       61.9  2.36 4090
       71.0 2.41 4480
       80   2.47 5270];
% Kormer data
%       P    T
dataK=[30  1895
       30  2079
       31.9  1895
       31.9  2079
       39.2  2447
       39.2  2553];

ph=logspace(3,log10(8.5e4),30);
[rhoh,Th,Eh]=CalcWaterHug(sp_IAPWS,ph);
[rhoh2,Th2,Eh2]=CalcWaterHug(sp_G_IAPWS_modified,ph);
np=10;
pc=linspace(ph(1),ph(end),np);
dz=water.data{5}.z;
dz(1:4)=dz(1:4)/1e1;
%
% Here is a plot of Hugoniot data and predictions from the modified model
figure(500)
sb1=subplot(211);
plot(ph/1e3,rhoh,'k--','LineWidth',2)
hold on
plot(ph/1e3,rhoh2,'k-','LineWidth',2)
plot(water.data{5}.P,1e3*water.data{5}.z,'ko','MarkerFaceColor','k','MarkerSize',8)
hold off
set(sb1,'FontSize',14)
xlabel('Pressure (GPa)','FontSize',20)
ylabel('Density (kg m^-^3)','FontSize',20)
text(40,1975,'IAPWS-95','FontSize',16,'Rotation',15)
text(20,1925,'Current','FontSize',16,'Rotation',25)
% 

sb2=subplot(212);
 plot(ph/1e3,Th,'k-','LineWidth',2)
 hold on
  errorbar(dataK(:,1),dataK(:,2),0.1*dataK(:,2),'ko','MarkerFaceColor','k','MarkerSize',8)
  errorbar(dataL(:,1),dataL(:,3),0.1*dataL(:,3),'k^','MarkerFaceColor','k','MarkerSize',8)
 hold off
set(sb2,'FontSize',14)
xlabel('Pressure (GPa)','FontSize',20)
ylabel('Temperature (K)','FontSize',20)


%
%calculate deviations between Decremps et al and the modified EOS
Ts=water.data{2}.T;
vels=1e3*water.data{2}.z;
id723=find(Ts==723);
id673=find(Ts==673);
id573=find(Ts==573);
id473=find(Ts==473);
id400=find(Ts==400);
Ps=water.data{2}.P;
% per Abramson and Brown figure 1, there is a systematic difference in the
% temperature calibration of pressure determinations between UW and the group in France.  
% This difference in calibration based on the figure is applied here:
Ps(id723)=Ps(id723)+.12;
Ps(id673)=Ps(id673)+.09;
Ps(id573)=Ps(id573)+.06; 
Ps(id473)=Ps(id473)+.03;
Ps(id400)=Ps(id400)+.02;
[~,velcs]=fnGval(sp_G_IAPWS_modified,[1e3*Ps(:) Ts(:)]);
devs=1e2*(vels-velcs)./velcs;
devs=1e2*(vels-velcs)./velcs;

%calculate deviations between Abramson and Brown and the modified EOS
Pab=water.data{1}.P;
Tab=water.data{1}.T;
velab=1e3*water.data{1}.z;
[~,velcab]=fnGval(sp_G_IAPWS_modified,[1e3*Pab(:) Tab(:)]);
devab=1e2*(velab-velcab)./velcab;

%calculate deviations between Ashara et al and the modified EOS
Pa=water.data{3}.P;
Ta=water.data{3}.T;
vela=1e3*water.data{3}.z;
[~,velca]=fnGval(sp_G_IAPWS_modified,[1e3*Pa(:) Ta(:)]);
deva=1e2*(vela-velca)./velca;

%
scrsz=get(0,'ScreenSize');
figsz=[1 .8*scrsz(4) 2*scrsz(3)/4 .8*scrsz(4)];
h=figure('Name','Water Sound Speeds','Position',figsz);
%
sb1=subplot(211);
Pp=PTd{1}/1e3; % use GPa here
%plot sound speeds
plot(Ps(id400),vels(id400),'ko','MarkerSize',6,'MarkerFaceColor','w')
hold on
plot(Pa,vela,'kd','MarkerSize',10,'MarkerFaceColor',[.7 .7 .7])
plot(Pab,velab,'ks','MarkerSize',10,'MarkerFaceColor','k')

plot(Pp,mask(:,[27 39 60 76])'.*velc(:,[27 39 60 76])','k','LineWidth',2)
hold off
xlabel('Pressure (GPa)','Fontsize',20)
ylabel('Temperature (K)','Fontsize',20)
axis([0 30 1000 8000])
set(sb1,'FontSize',16);

sb2=subplot(212);
plot(Ps,devs,'ko','MarkerSize',5,'MarkerFaceColor','k')
hold on
plot(Pa,deva,'kd','MarkerSize',8,'MarkerFaceColor',[.7 .7 .7])
plot(Pab,devab,'ks','MarkerSize',8,'MarkerFaceColor','w')
axis([1 30 -8 8])
hold off
xlabel('Pressure (GPa)','FontSize',20)
ylabel('(Obs-Calc)/Obs (%)','FontSize',20)
set(sb2,'FontSize',16)
