% This script accompanies the paper "Local Basis Function Representations of Thermodynamic Surfaces:
%                                     Water at High Pressure and
%                                     Temperature as an Example"
% the following matlab sections are associated with the section of the paper titled:
%                                    "Reanalysis of Sound Speed
%                                    Measurements"
% although the entire script can be executed at once, it is best to run
% section by section after reading the embedded comments

% SECTION ONE
% This section does calculations necessary to make Figure 1

LT=LT_data;  % this function returns the LT data set.

% create a grid of IAPWS sound speeds
P=0.1:10:434.1;
T=250:10:480;
%call local basis function representation for IAPWS
[~,vel]=fnIAPWSval({P,T}); % since only sound speeds are required, output that preceeds sounds speeds not preserved
% the full call to fnIAPWS is:  [rho,vel,G,Cp,alpha,S,U,H,K,Kp,mask]=fnIAPWSval(input,sp)
% The alternate (slower) method is to use the original IAPWS formulation:
% (commented out here);
% [rho,vel,G,Cp,alpha,S,U,H,Cv,Kt,gamma,Pout]=IAPWS95({P,T},'P')

% approximate melting phase lines from literature
Tmelt=[273 252 257 261 266];
Pmelt=[0.1 226 mean([226 350]) 350 450];
nPm=length(Tmelt);
[Pm,Tm]=meshgrid(P,T);
Vmelt=interp2(Pm,Tm,1.01*vel',Pmelt,Tmelt);

% make Figure 1s
figure1 = figure;
figure1.Name='Figure 1 Sound Speeds for Water';
figure1.NumberTitle='off';
figure1.Position=[111 136 709 669];
clf
% Create axes
axes1 = axes('Parent',figure1);
plot3(LT(:,1),LT(:,2),LT(:,3),'ko','MarkerFaceColor','k','MarkerSize',10)
hold on
plot3(Pmelt,Tmelt,Vmelt,'k-','LineWidth',4)
surf(P,T,vel','SpecularExponent',1,...
    'FaceColor',[.6 .6 .6],...
     'EdgeColor',[.6 .6 .6]) 
hold off
box 'on'
axis([ 0 430 250 490 1000 2300])
xlabel('Pressure (MPa)','FontSize',20,'rotation',20)
ylabel('Temperature (K)','FontSize',20,'rotation',-30)
zlabel('Sound Speed (m/s)','FontSize',20)
set(axes1,'FontSize',16);
grid on
hold off

%%
% SECTION TWO
%Make Figure 2

% set up parameters for spline fit to LT data
Pc=linspace(0,500,50);   % control points in pressure for sound speed fit
Tc=linspace(240,500,50); % control points in temperature for sound speed fit
RegFac=[2 2];  % add two regularization points for every grid point
Ordr=[6 6];  % order is adjustable and connected to interval spacing - larger intervals need more degrees of freedom
mdrv=[2 2];   % second derivative is being minimized
lam=65*[1 1];  % lam chosen to achieve chi sqr ~ 1

sp_LTFit=spdft(LT(:,1:2),LT(:,3),LT(:,4),{Pc,Tc},lam,RegFac,Ordr,mdrv,[]);  % call to spline fitting function
%sp_LTFit.Data  % if uncommented this displays the statistics of fit in command window

% This evaluates IAPWS at Lin and Trusler data sites
[~,velid]=fnIAPWSval(LT(:,1:2));

%This returns sound speeds at data sites per the polynomial reported in Lin and Trusler 
velLTF=returnLTv(LT(:,1:2));

% This returns sound speeds per the local basis function fit
velLTMF=fnval(sp_LTFit,LT(:,1:2)');

% calculate percentage deviations of the Lin and Trusler sound speeds from
% the three representations
devsLTi=1e2*((LT(:,3)-velid)./(LT(:,3)));
devsLTF=1e2*((LT(:,3)-velLTF)./(LT(:,3)));
devsLTMF=1e2*((LT(:,3)-velLTMF')./(LT(:,3)));

% make the plots
figure2=figure;
figure2.Name='Figure 2 Sound Speed Deviations';
figure2.NumberTitle='off';
figure2.Position=[100 100 1600 500];
sp1=subplot(131);
plot3(LT(:,1),LT(:,2),devsLTi,'ko','MarkerFaceColor','k','MarkerSize',8)
hold on
surf([0 0 500 500 0],[250 500 500 250 250],zeros(5,5),'FaceColor',[0.7 0.7 0.7],...
    'EdgeColor',[1 1 1],...
    'AlphaData',0.3)
xlabel('Temperature (K)','FontSize',20,'Rotation',20)
ylabel('Pressure (MPa)','FontSize',20,'Rotation',-10)
zlabel('Sound Speed Deviations (%)','FontSize',20)
axis([0 500 250 500 -.4 .4])
view([-50 8])
set(sp1,'FontSize',14);
hold off
title('Deviations from IAPWS')
grid on
box on

sp2=subplot(132);
plot3(LT(:,1),LT(:,2),devsLTF,'ko','MarkerFaceColor','k','MarkerSize',8)
hold on
surf([0 0 500 500 0],[250 500 500 250 250],zeros(5,5),'FaceColor',[0.7 0.7 0.7],...
    'EdgeColor',[1 1 1],...
    'AlphaData',0.3)
axis([0 500 250 500 -.4 .4])
xlabel('Temperature (K)','FontSize',20,'Rotation',20)
ylabel('Pressure (MPa)','FontSize',20,'Rotation',-10)
view([-50 8])
zlabel('Sound Speed Deviations (%)','FontSize',20)
set(sp2,'FontSize',14);
hold off
title('Deviations from Polynomial Fit')
box on
grid on

sp3=subplot(133);
plot3(LT(:,1),LT(:,2),devsLTMF,'ko','MarkerFaceColor','k','MarkerSize',8)
hold on
surf([0 0 500 500 0],[250 500 500 250 250],zeros(5,5),'FaceColor',[0.7 0.7 0.7],...
    'EdgeColor',[1 1 1],...
    'AlphaData',0.3)
axis([0 500 250 500 -.4 .4])
xlabel('Temperature (K)','FontSize',20,'Rotation',20)
ylabel('Pressure (MPa)','FontSize',20,'Rotation',-10)
view([-50 8])
zlabel('Sound Speed Deviations (%)','FontSize',20)
set(sp3,'FontSize',14);
title('Deviations from Spline Fit')
hold off
box on
grid on


%%
% SECTION THREE
% Make Figure 3a
% construct a control point grid for the representation of Gibbs energy:
PTc={linspace(0,400.1,40),linspace(250,475,30)};

% construct a data grid for problem
Pd=0.1:10:400.1; % data points in P (used for integration in P)  (defined in MPa)
Td=logspace(log10(252),log10(474),400); % the number of points can be adjusted to find requisit number for numerical stability
Td(1)=252; Td(end)=474; % the logspace function returns end points that may deviate by machine precision from those specified - this forces them to match
PTd={Pd',Td'};
% control grid in temperature for the "eqst" integration:
Tc=linspace(252,474,80); % control point grid in T (used to determine V(T) spline knots ) vary to test numerical stability
% IAPWS evaluated on the grid - defines the 1 bar starting point
[rhoi,veli,Gi,Cpi,~,~,~,~,~,~,mask]=fnIAPWSval(PTd);
%This is the Lin and Trusler polynomial evaluated on the specified grid 
velLTF=returnLTv(PTd);

% This evaluates the local basis function representation of the Lin and Trusler sound speeds on the specified grid
velLTMF=fnval(sp_LTFit,PTd);

% Set parameter that control calculations using eqst
lamv=1e0;  % degree of smoothing - always useful to start with default of 1
lam_fac=1.2; % this allows more smoothing at higher pressures (multiply prior lamv by this at each pressure step)
pflg=0; % print flag set to 1 to watch the progress of the predictor-corrector algorithm - if things fail

% here is the integration
[rhoiI,CpiI,GiI]=eqst(PTd,Tc,rhoi(1,:),Cpi(1,:),veli,Gi(1,:),mask,lamv,lam_fac,pflg);
%
% Convert G, rho, and Cp into a single local basis function representation
% for G - need to set parameters:

wt=1e0*[2e3 1e0];  % weights for rho and Cp => later versions will correct this so that 1 will be the natural values
lamg=1e0*[1 1e-1]; % damping in P and T for the regularization of the fit
mdrv=[5 5];    % derivatives of G to make small
ordr=[8 8];    % order of spline in P and T
nreg=1;
normflg=1;
sp_GiI=spgft(GiI(1,:),rhoiI,CpiI,wt,PTd,PTc,mask,lamg,mdrv,ordr,nreg,normflg) ; % do the work

% Evaluate properties of the new representation on the grid 
[rhociI,velciI,GciI,CpciI]=fnGval(sp_GiI,PTd);  
% plot deviations of representation to IAPWS:
txt1='Density Deviations (ppm)';
txt2='Specific Heat Deviations (%)';
txt3='Expanded Scale';
np=length(Pd);
pid=1:10:41; % plot .1 100 200 300 and 400 MPa results
%
figure3=figure;
figure3.Name='Figure 3 Water EOS Deviations';
figure3.NumberTitle='off';
figure3.Position=[100 100 1400 700];
subplot1=subplot(231);
plot(Td,1e6*(rhociI(pid(1),:)-rhoi(pid(1),:))/1000,'k-','LineWidth',2)
hold on
plot(Td,1e6*(rhociI(pid(2),:)-rhoi(pid(2),:))/1000,'k--','LineWidth',2)
plot(Td,1e6*(rhociI(pid(3),:)-rhoi(pid(3),:))/1000,'k:','LineWidth',2)
plot(Td,1e6*(rhociI(pid(4),:)-rhoi(pid(4),:))/1000,'k-.','LineWidth',2)
plot(Td(1:8:end),1e6*(rhociI(pid(5),1:8:end)-rhoi(pid(5),1:8:end))/1000,'ko')
hold off
xlabel('Temperature (K)','FontSize',20)
ylabel('(obs-IAPWS)/1000 (ppm)','FontSize',20)
title('IAPWS Control','FontSize',20)
axis([250 475 -30 30])
set(subplot1,'FontSize',14);
text(300,25,txt1,'FontSize',16)
text(300,21,txt3,'FontSize',16)

subplot2=subplot(234);
plot(Td,1e2*(CpciI(pid(1),:)-Cpi(pid(1),:))/4000,'k-','LineWidth',2)
hold on
plot(Td,1e2*(CpciI(pid(2),:)-Cpi(pid(2),:))/4000,'k--','LineWidth',2)
plot(Td,1e2*(CpciI(pid(3),:)-Cpi(pid(3),:))/4000,'k:','LineWidth',2)
plot(Td,1e2*(CpciI(pid(4),:)-Cpi(pid(4),:))/4000,'k-.','LineWidth',2)
plot(Td(1:8:end),1e2*(CpciI(pid(5),1:8:end)-Cpi(pid(5),1:8:end))/4000,'ko')
xlabel('Temperature (K)','FontSize',20)
ylabel('(obs-IAPWS)/4000 (%)','FontSize',20)
axis([250 475 -.2 .2])
set(subplot2,'FontSize',14);
text(300,-.15,txt2,'FontSize',16)
text(300,-.18,txt3,'FontSize',16)
hold off

% Make Figure 3b
% Integrate the Lin and Trusler polynomial for sound speeds
[rhoLT,CpLT,GLT]=eqst(PTd,Tc,rhoi(1,:),Cpi(1,:),velLTF,Gi(1,:),mask,lamv,lam_fac,pflg);

% determine the representation of Gibbs energy
sp_GLT=spgft(GLT,rhoLT,CpLT,wt,PTd,PTc,mask,lamg,mdrv,ordr,nreg,normflg);

%evaluate properties from this Gibbs representation
[rhocLT,velLT,GcLT,CpcLT]=fnGval(sp_GLT,PTd);  % this takes MPa

% Plot deviations relative to IAPWS
subplot3=subplot(232);
plot(Td,1e6*(rhocLT(pid(1),:)-rhoi(pid(1),:))/1000,'k-','LineWidth',2)
hold on
plot(Td,1e6*(rhocLT(pid(2),:)-rhoi(pid(2),:))/1000,'k--','LineWidth',2)
plot(Td,1e6*(rhocLT(pid(3),:)-rhoi(pid(3),:))/1000,'k:','LineWidth',2)
plot(Td,1e6*(rhocLT(pid(4),:)-rhoi(pid(4),:))/1000,'k-.','LineWidth',2)
plot(Td(1:8:end),1e6*(rhocLT(pid(5),1:8:end)-rhoi(pid(5),1:8:end))/1000,'ko')
hold off
title('Lin and Trussler Fit','FontSize',20)
xlabel('Temperature (K)','FontSize',20)
ylabel('(obs-IAPWS)/1000 (ppm)','FontSize',20)
axis([250 475 -300 300])
set(subplot3,'FontSize',14);
text(300,250,txt1,'FontSize',16)


subplot4=subplot(235);
plot(Td,1e2*(CpcLT(pid(1),:)-Cpi(pid(1),:))/4000,'k-','LineWidth',2)
hold on
plot(Td,1e2*(CpcLT(pid(2),:)-Cpi(pid(2),:))/4000,'k--','LineWidth',2)
plot(Td,1e2*(CpcLT(pid(3),:)-Cpi(pid(3),:))/4000,'k:','LineWidth',2)
plot(Td,1e2*(CpcLT(pid(4),:)-Cpi(pid(4),:))/4000,'k-.','LineWidth',2)
plot(Td(1:8:end),1e2*(CpcLT(pid(5),1:8:end)-Cpi(pid(5),1:8:end))/4000,'ko')
xlabel('Temperature (K)','FontSize',20)
ylabel('(obs-IAPWS)/4000 (%)','FontSize',20)
axis([250 475 -2 2])
set(subplot4,'FontSize',14);
text(300,-1.5,txt2,'FontSize',16)
hold off

% Make Figure 3c
% Integrate the local basis function representation for sound speeds
[rhoLTM,CpLTM,GLTM]=eqst(PTd,Tc,rhoi(1,:),Cpi(1,:),velLTMF,Gi(1,:),mask,lamv,lam_fac,pflg);

% convert to representation of Gibbs energy
sp_GLTM=spgft(GLTM,rhoLTM,CpLTM,wt,PTd,PTc,mask,lamg,mdrv,ordr,nreg,normflg);

%evaluate properties from this Gibbs representation
[rhocLTM,velLTM,GcLTM,CpcLTM,alpha2]=fnGval(sp_GLTM,PTd);  

% plot deviations:
subplot5=subplot(233);
plot(Td,1e6*(rhocLTM(pid(1),:)-rhoi(pid(1),:))/1000,'k-','LineWidth',2)
hold on
plot(Td,1e6*(rhocLTM(pid(2),:)-rhoi(pid(2),:))/1000,'k--','LineWidth',2)
plot(Td,1e6*(rhocLTM(pid(3),:)-rhoi(pid(3),:))/1000,'k:','LineWidth',2)
plot(Td,1e6*(rhocLTM(pid(4),:)-rhoi(pid(4),:))/1000,'k-.','LineWidth',2)
plot(Td(1:8:end),1e6*(rhocLTM(pid(5),1:8:end)-rhoi(pid(5),1:8:end))/1000,'ko')
hold off
title('Gibbs Energy Spline','FontSize',20)
xlabel('Temperature (K)','FontSize',20)
ylabel('(obs-IAPWS)/1000 (ppm)','FontSize',20)
axis([250 475 -300 300])
set(subplot5,'FontSize',14);
text(300,250,txt1,'FontSize',16)


subplot6=subplot(236);
plot(Td,1e2*(CpcLTM(pid(1),:)-Cpi(pid(1),:))/4000,'k-','LineWidth',2)
hold on
plot(Td,1e2*(CpcLTM(pid(2),:)-Cpi(pid(2),:))/4000,'k--','LineWidth',2)
plot(Td,1e2*(CpcLTM(pid(3),:)-Cpi(pid(3),:))/4000,'k:','LineWidth',2)
plot(Td,1e2*(CpcLTM(pid(4),:)-Cpi(pid(4),:))/4000,'k-.','LineWidth',2)
plot(Td(1:8:end),1e2*(CpcLTM(pid(5),1:8:end)-Cpi(pid(5),1:8:end))/4000,'ko')
xlabel('Temperature (K)','FontSize',20)
ylabel('(obs-IAPWS)/4000 (%)','FontSize',20)
axis([250 475 -2 2])
set(subplot6,'FontSize',14);
text(300,-1.5,txt2,'FontSize',16)
hold off