function [rho,Cp,G,spv]=eqst(PT,Tc,rhoin,Cpin,vel,Go,mask,lam,lam_fac,pflg)
% Usage:
%     [rho,Cp,G,spv]=eqst(PT,Tc,rhoin,Cpin,vel,Go,mask,lam,lam_fac,pflg)
%      PT is cell defining the P and T grid
%      Tc defines control points for spline fit to volumes vs temperature
%      rhoin is vector of densities at base of isotherm, 
%      Cpin is vector of specific heats at base of isotherm or grid over
%              all pressures and temepratures - to force the Cp values
%      vel is a grid of sound speeds
%      Go is the vector of Gibbs energy at base of isotherm
%      mask is a grid of ones or Nan's to define the physical regime for the integration
%      lam is the damping at the base of the isotherms
%      lam_fac is the factor increasing lam for each step in pressure -
%              pragmatic discovery is that more damping is required as the
%                 integration is pushed to high pressures
%      pflg=1 plots quantities associated with the adiabatic to isothermal
%      correction - provides information if the integration is not working
%
%      This algorithm can suffer from the growth of numerical instabilities
%      at higher pressures.  Two stabilizing ideas are incorporated. 
%            (1) with each step in pressure, the damping of V(T) fits is
%            increased by a specified factor "lam_fac".
%                 (Try 1 as the default and increase slowly if that does not work). 
%            (2) Second derivatives of V(T) are averaged over 5 points 
%                to reduce growth of point to point instabilities
%       If it remains difficult to stabilize the solution,
%       setting  "pflg" to 1 plots quantities at each pressure step.  This
%       highlights at what pressure and over what temperature range the
%       algorithm fails. This can then provide guidance in modification of
%       the input sound speed surface.  A defect in the sound speed surface 
%       is a typical source of integration difficulty.
%       
% use MKS units (P in MPa)i
%  JMB 2014

% emperical choices for the spline fit in temperature - seems to work in most
% cases:
ordr=8;
mdrv=6;
nReg=1;

kT=ordr;
P=PT{1};
T=PT{2};
T=T(:)'; % enforce T as row
nT=length(T);
nTc=length(Tc);
if(nTc>nT),error('Too many control points relative to data points for temperatures'),end
facP=1e6;  %conversion of P from MPa to Pa;
P=facP*P(:); % insist that P is a column and in Pa 
nP=length(P);

velsq= (vel).^(-2);
% test whether the input velocities are finite and reasonably bounded
sizenan=length(find(isnan(vel)));
sizenotfinite=length(find(vel<100 | vel>4e4) | ~isreal(vel));
%if(sizenan~=0 | sizenotfinite~=0), error('velocity matrix contains bad points'),end
spv=csapi({P, T},velsq); % assumption that vel is defined everywhere on the grid in P and T. 
%                           csapi is a MATLAB provided cubic spline interpolations function
velsqInt=fnval(fnder(spv,[-1 0]),{P,T}); % analytically evaluate the integral of the spline representing 1/vel^2
dvelsqInt=diff(velsqInt);

Tr=mkgrid(Tc,nReg);
nTr=length(Tr);
nv=min(size(Cpin));
if(nv==1),fit_Cp=1;else fit_Cp=0;end
NA=sqrt(nTc);
wtR=area_wt(Tr);

% set up b spline for V(T)
Tknts=optknt(Tc,ordr);
Tcol = spcol(Tknts,ordr,brk2knt(T,ordr));
Vcol=Tcol(1:ordr:end,:);
d2Vcol=Tcol(3:ordr:end,:);
Tcolreg = spcol(Tknts,ordr,brk2knt(Tr,ordr));
RegCol=Tcolreg((mdrv+1):ordr:end,:);  % specified derivative of V with T to be made small
A=[Vcol;lam*NA*(wtR(:)*ones(1,nTc)).*RegCol];

sp.form='B-';
sp.knots=Tknts;
sp.number=length(Tc);
sp.order=kT;
sp.dim=1;


rho=zeros(nP,nT);
alpha=rho;
Cp=rho;
G=rho;
cor=rho;
dVdT=rho;
d2VdT2=rho;

%set fit flag for specific heat 
if fit_Cp
        Cp(1,:)=Cpin(:)';
else
    Cp=Cpin; 
end

rho(1,:)=rhoin(:)';
G(1,:)=Go(:)';
V=rho(1,:).^(-1);
B=[V(:);zeros(nTr,1)];
sp.coefs=(A\B)'; % solve for V(T)

dVdT(1,:)=fnval(fnder(sp),T);
d2VdT2(1,:)=fnval(fnder(sp,2),T);
alpha(1,:)=dVdT(1,:).*rho(1,:);
cor(1,:)=alpha(1,:).^2./Cp(1,:).*T;
incld=find(not(isnan(mask(1,:))));

% plot quantities if debugging
if pflg
    figure(100)
    txt=sprintf('P= %6.0f ',P(1)/1e6);
    subplot(411)
    plot(T(incld),rho(1,:))
      title('Density')
    rm=max(rho(1,:));
    text(300,rm-10,txt)
     subplot(412)
    plot(T(incld),Cp(1,:),'-o')
      title('Cp')
    subplot(413)
    plot(T(incld),d2VdT2(1,:),'-o')
      title('d2VdT2')
    subplot(414)
    plot(T(incld),cor(1,:))
    title('Correction')
    pause
end

% begin integration to high pressure - starting at the second pressure
% point
for i=2:nP
    icount=0;
    test=1;
    incld=find(not(isnan(mask(i,:))));
    lam=lam_fac*lam;
% increment densities using velocity integral and adiabatic correction term
    cor(i,incld)=.8*cor(i-1,incld);   % predict that the correction for next step is smaller than previous step 
while (test>1e-14 && icount<20); % stop when corrections do not change from iteration to iteration
    icount=icount+1;
    rho(i,incld)= rho(i-1,incld)+dvelsqInt(i-1,incld)+trapz(P(i-1:i),cor(i-1:i,incld)); % haven't found a stable way to remove this trapezoidal integration
    V=rho(i,incld).^(-1);
    A=[Vcol(incld,:); lam*NA*(wtR(:)*ones(1,nTc)).*RegCol];
    B=[V(:); zeros(nTr,1)];   
    sp.coefs=(A\B)';
    dVdT(i,incld)=fnval(fnder(sp),T(incld));
    d2VdT2(i,incld)=fnval(fnder(sp,2),T(incld));   
    alpha(i,(incld))=dVdT(i,(incld)).*rho(i,(incld));
    if fit_Cp  % section to updata Cp
%following does trapezoidal integration for Cp at new pressure
        Cp(i,incld)=Cp(i-1,incld) - trapz(P(i-1:i),(ones(2,1)*T(incld)).*d2VdT2(i-1:i,incld));
% following is a 5 point averaging of Cp vs T to damp high frequency numerical instabilities
        for j=3:(length(incld)-2)
            Cp(i,incld(j))=mean(Cp(i,incld(j-2:j+2)));
        end       
    end  % fit CP
% Calculate new value of correction term
    corN=alpha(i,incld).^2./Cp(i,incld).*T(incld);
    test=max(abs(cor(i,incld)-corN));
    cor(i,incld)=corN;
% plot stuff if debugging:    
  if pflg
    figure(100)
    txt=sprintf('P= %6.0f ',P(i)/1e6);
    subplot(411)
    plot(T(incld),rho(i,incld))
      title('Density')
    rm=max(rho(i,incld));
    text(300,rm-10,txt)
     subplot(412)
    plot(T(incld),Cp(i,incld),'-o')
      title('Cp')
    subplot(413)
    plot(T(incld),d2VdT2(i,incld),'-o')
      title('d2VdT2')
    subplot(414)
    plot(T(incld),trapz(P(i-1:i),cor(i-1:i,incld)),T(incld), dvelsqInt(i-1,incld),'--')
    title('Correction')
    pause
  end   % end of plotting

end   %end of interations on correction term
G(i,incld)=G(i-1,incld)+ trapz(P(i-1:i),rho(i-1:i,incld).^-1);
end  % end of pressures loop

 
 function wt=area_wt(PT)
% this subfunction assigns a weight to every regularization point based on
% the volume encompassed by that point
if(iscell(PT))
 if (length(PT)==2)   
    p=PT{1};
    dp=diff(p(:))';
    t=PT{2};
    dt=diff(t(:))';
    wtp=.5*[dp(1) dp(1:end-1)+dp(2:end) dp(end)];
    wtt=.5*[dt(1) dt(1:end-1)+dt(2:end) dt(end)];
    wt=wtp(:)*wtt;
    wtsum=sum(wt(:));
    wt=sqrt(wt/wtsum);
 elseif( length(PT)==3)  
      p=PT{1};
      dp=diff(p(:))';
      t=PT{2};
      dt=diff(t(:))';
      m=PT{3};
      dm=diff(m(:))';
    wtp=.5*[dp(1) dp(1:end-1)+dp(2:end) dp(end)];
    wtt=.5*[dt(1) dt(1:end-1)+dt(2:end) dt(end)];
    wtm=.5*[dm(1) dm(1:end-1)+dm(2:end) dm(end)];
    [a,b,c]=ndgrid(wtp,wtt,wtm);
    wt=a.*b.*c;
    wtsum=sum(wt(:));
    wt=sqrt(wt/wtsum);
  end
 else    
    dx=diff(PT(:))';
    wtx=.5*[dx(1) dx(1:end-1)+dx(2:end) dx(end)];
    wtsum=sum(wtx(:));
    wt=sqrt(wtx/wtsum);
end
 
function xg=mkgrid(x,nreg)
% this subfunction puts nreg points between every value of x
xg=x(:)';
dx=diff(xg);
if nreg>1;
for i=2:nreg
    xg=[xg x(1:end-1)+(i-1)*dx/nreg];
end
end
xg=sort(xg);

